# pylint: skip-file
import unittest.mock

import ioc
from unimatrix.ext import crypto
from unimatrix.lib.datastructures import DTO

import frf
from frf.lib.test import ProtectedResourceTestCase
from .base import ProtectedResource


class BearerTokenSignatureTestCase(ProtectedResourceTestCase):
    base_path = 'test'
    endpoints = [
        (base_path, ProtectedResource)
    ]

    def get_asgi_application(self):
        return frf.get_asgi_application(allowed_hosts=['*'])

    def test_bearer_with_valid_signature(self):
        self.refresh_token()
        response = self.request(self.client.get, self.base_path + '/me')
        self.assertEqual(response.status_code, 200, response.json())

    def test_bearer_with_invalid_signature(self):
        self.signer.key = crypto.SecretKey({'secret': b'sfsdfs'})
        self.refresh_token(secret='foo') # nosec
        response = self.request(self.client.get, self.base_path + '/me')
        self.assertEqual(response.status_code, 403, response.json())



import uvicorn
from frf import AnonymousResource
from frf import get_asgi_application


class Thing(AnonymousResource):

    async def retrieve(self, request, resource_id: int):
        return {

            "id": resource_id,
            "message": "Hello world!"
        }

    async def list(self):
        return [
            {"message": "First item in a list of Thing!"},
            {"message": "Second item in a list of Thing!"},
        ]


app = get_asgi_application(allowed_hosts=['*'])
app.add_resource(Thing)

if __name__ == '__main__':
    uvicorn.run(app, host="127.0.0.1", port=8000)



"""Declares :class:`OIDCProvider`."""
import fastapi
import frf


class TokenEndpoint(frf.AnonymousResource):
    default_provider = 'default'

    async def list(self):
        return await self.retrieve('default')

    @frf.action(name='token', detail=True, methods=['post'])
    async def token(self, resource_id: str, request: fastapi.Request):
        pass



if __name__ == '__main__':
    import uvicorn
    app = frf.get_asgi_application(allowed_hosts=["*"])
    app.add_resource(TokenEndpoint, '')
    uvicorn.run(app, host="127.0.0.1", port=8000)



# pylint: skip-file


def constant_time_compare(val1, val2):
    """Returns True if the two strings are equal, False otherwise.
    The time taken is independent of the number of characters that match.
    For the sake of simplicity, this function executes in constant time only
    when the two strings have the same length. It short-circuits when they
    have different lengths.
    """
    val1 = str.encode(val1)
    val2 = str.encode(val2)
    if len(val1) != len(val2):
        return False
    result = 0
    for x, y in zip(val1, val2):
        result |= x ^ y
    return result == 0




# pylint: skip-file
from bearer import *


class BearerSubjectResolver(frf.BearerSubjectResolver):

    async def verify_signature(self, token):
      # Be aware that the signature of the Bearer token is not verified
      # yet, so the claims should not be trusted.
      signing_key = lookup_subject_key(token.get_unverified('sub'))
      return token.verify(signing_key=signing_key)


if __name__ == '__main__':
    frf.provide('SubjectRepository', SubjectRepository())
    frf.provide('BearerSubjectResolver', BearerSubjectResolver(), force=True)

    uvicorn.run('auth:app', host="127.0.0.1", port=8000, reload=True)



# pylint: skip-file
import ioc

import frf
from frf.lib.test import ProtectedResourceTestCase
from .base import ProtectedResource


class ProtectedResource(ProtectedResource):
    required_scopes = ['foo', 'bar']


class BearerTokenScopesTestCase(ProtectedResourceTestCase):
    base_path = 'test'
    endpoints = [
        (base_path, ProtectedResource)
    ]

    def get_asgi_application(self):
        return frf.get_asgi_application(allowed_hosts=['*'])

    def test_list_returns_200_with_correct_scopes_equal(self):
        self.refresh_token(scopes=['foo', 'bar'])
        response = self.request(self.client.get, self.base_path)
        self.assertEqual(response.status_code, 200, response.json())

    def test_list_returns_200_with_correct_scopes_superset(self):
        self.refresh_token(scopes=['foo', 'bar', 'baz'])
        response = self.request(self.client.get, self.base_path)
        self.assertEqual(response.status_code, 200, response.json())

    def test_list_returns_403_with_invalid_scope_subset(self):
        self.refresh_token(scopes=['foo'])
        response = self.request(self.client.get, self.base_path)
        self.assertEqual(response.status_code, 401, response.json())

    def test_list_returns_403_with_invalid_scope_missing(self):
        self.refresh_token()
        response = self.request(self.client.get, self.base_path)
        self.assertEqual(response.status_code, 401, response.json())

    def test_list_returns_403_with_invalid_scope_disjoint(self):
        self.refresh_token(scopes=['taz'])
        response = self.request(self.client.get, self.base_path)
        self.assertEqual(response.status_code, 401, response.json())



"""Declares various viewset implementations."""
from frf.views import APIView


class GenericViewSet(APIView):
    """A generic viewset implementation and base class for all viewsets. A
    :class:`GenericViewSet` represents a RESTful API resource for which
    operation may be defined. It has no endpoints by itself - the
    :class:`GenericViewSet` class must be subclassed in order to create a
    RESTFUL API resource.

    Endpoints are created by implementing any of the :meth:`create`,
    :meth:`list`, :meth:`retrieve`, :meth:`update` or :meth:`delete` methods
    on a subclass of :class:`frf.GenericViewSet`.
    """
    __module__ = 'frf'



"""Declares decorators to use with FastAPI RESt Framework."""
import functools

import ioc
import ioc.loader
from fastapi.responses import HTMLResponse


__all__ = [
    'action',
    'response_model',
    'spa',
]


class action:
    """Marks extra methods for routing."""

    def __init__(self, detail=False, name=None, methods=None):
        self.detail = detail
        self.name = name
        self.methods = methods or ['get']

    def __call__(self, func):
        self.name = self.name or func.__name__
        self.func = func
        func.action = self
        return func

    def add_to_app(self, app, base_path, cls):
        """Add the action to the FastAPI application."""
        base_path = str.rstrip(base_path, '/')
        if self.detail:
            base_path = f'{base_path}/{{resource_id}}'
        for method in self.methods:
            cls.as_method_handler(app, f'{base_path}/{self.name}',
                method, self.func.__name__)


def response_model(cls):
    """Decorates a request handler to use the given response model `cls`."""
    def decorator(func):
        func.response_model = cls
        return func
    return decorator


def spa(template, requires='TemplateService', path=None, detail=False):
    """Decorate a request handler to indicate that it returns a Single Page
    Application (SPA). The return value of the handler is expected to be
    a dictionary holding a template rendering context.
    """

    def decorator_factory(func):
        @functools.wraps(func)
        @action(name=path or func.__name__, detail=detail, methods=['get'])
        async def decorated(self, *args, **kwargs):
            svc = ioc.require(requires)
            ctx = await func(self, *args, **kwargs)
            ctx['kwargs'] = kwargs
            return HTMLResponse(await svc.render_to_string(template, ctx))

        decorated.response_class = HTMLResponse
        return decorated

    return decorator_factory



# pylint: skip-file
import frf


class ProtectedResource(frf.Resource):

    async def list(self):
        return {}

    @frf.action(name='me')
    async def current_user(self):
        return {
            'sub': self.claims['sub']
        }



"""Declares :class:`ResourceEndpointTestCase`."""
import copy
import unittest

import unimatrix.ext.jose.__ioc__
from unimatrix.ext import jose
from unimatrix.ext.jose.tests import signer
from fastapi.testclient import TestClient


class ResourceEndpointTestCase(unittest.TestCase):
    """Provides boilerplate to test :class:`~frf.Resource` objects."""
    endpoints = []

    def setUp(self):
        # TODO: Hack
        unimatrix.ext.jose.__ioc__.setup_ioc()
        self.asgi = self.get_asgi_application()
        for path, cls in self.endpoints:
            self.asgi.add_resource(cls, base_path=path)
        self.client = TestClient(self.asgi)

    def request(self, func, *args, **kwargs):
        headers = kwargs.pop('headers', None) or {}
        return func(*args, headers={**headers, **copy.deepcopy(self.headers)},
            **kwargs)

    def get_headers(self):
        """Return the default headers."""
        return {}

    def get_asgi_application(self):
        """Return the :class:`~frf.ASGIApplication` instance that is used
        during this test. Subclasses must override this method.
        """
        raise NotImplementedError


class ProtectedResourceTestCase(ResourceEndpointTestCase):
    audience = None
    scopes = None

    @property
    def headers(self):
        return {
            'Authorization': f"Bearer {bytes.decode(self.token)}"
        }

    def setUp(self):
        super().setUp()
        self.signer = self.get_signer()
        self.jwt, self.token = self.get_bearer_token()

    def get_signer(self):
        return signer.HMACSigner()

    def get_bearer_token(self):
        jwt = jose.jwt.sync({'aud': self.audience, 'sub': 'foo'},
            signer=self.signer)
        return jwt, bytes(jwt)

    def refresh_token(self, aud=None, claims=None, scopes=None, **kwargs):
        claims = {'sub': 'foo', **(claims or {})}
        if scopes is not None:
            claims['scopes'] = scopes
        if aud is not None:
            claims['aud'] = aud
        self.jwt = jose.jwt.sync(claims, signer=self.signer)
        self.token = bytes(self.jwt)



"""Declares :class:`ServiceMetadataResource`."""
from fastapi import Request

from ..dependency import Injected
from ..views import AnonymousResource


class ServiceMetadataResource(AnonymousResource):
    resource_name = "Metadata"

    async def list(self, request: Request, svc=Injected('APIMetadataService')):
        """Return a JSON object describing all API properties."""
        return await svc.get(self.asgi, request)



# pylint: skip-file
import ioc

from .services import APIMetadataService
from .services import HealthCheckService


def setup_ioc():
    ioc.provide('APIMetadataService', APIMetadataService())
    ioc.provide('HealthCheckService', HealthCheckService())
